/*    ********************************
      *      CONTR & BASIC           *
      *   UA6ANN  Alexander Kostyuk  *
      *   RV3BJ   Evgeny Fadeev      *
      ********************************
*/
/*
This source is based on BASIC interpreter source code 
called Small BASIC (another name is Tiny Basic) by Herbert Schildt 
in Turbo C: The Complete Reference. 
*/


/************************************************************************************
 $Workfile:  $
 $Header:    $
 $Log:       $
************************************************************************************/
#include "includes.h"

char __flash *bas_prog;  /* holds expression to be analyzed */
ARITH_BASE bas_variables[26];
char bas_token[64];
char bas_token_type;
char bas_tok;
LABEL bas_label_table[NUM_LAB];
int bas_ftos;  /* index to top of FOR stack */
int bas_gtos;  /* index to top of GOSUB stack */
FOR_STACK bas_fstack[FOR_NEST]; /* stack for FOR/NEXT loop */
char __flash *bas_gstack[SUB_NEST];	/* stack for gosub */

__flash char _print_[] =  "print";
__flash char _input_[] =  "input";
__flash char _if_[] =     "if";
__flash char _then_[] =   "then";
__flash char _goto_[] =   "goto";
__flash char _for_[] =    "for";
__flash char _next_[] =   "next";
__flash char _to_[] =     "to";
__flash char _gosub_[] =  "gosub";
__flash char _return_[] = "return";
__flash char _end_[] =    "end";
__flash char _empty_[] =   "";
__flash char   _adc_[] = "adc";
__flash char   _dac_[] = "dac";
__flash char   _setbit_[] = "setb";
__flash char   _clrbit_[] = "clrb";
__flash char   _tstbit_[] = "tstb";
__flash char   _delay_[] = "delay";
__flash char   _norm_[] = "scale";
__flash char   _rem_[] = "rem";
__flash commands_tbl table[] = { /* Commands must be entered lowercase */
  _print_, PRINT, /* in this table. */
  _input_, INPUT,
  _if_, IF,
  _then_, THEN,
  _goto_, GOTO,
  _for_, FOR,
  _next_, NEXT,
  _to_, TO,
  _gosub_, GOSUB,
  _return_, RETURN,
  _adc_,ADC,
  _dac_,DAC,
  _setbit_, SETBIT,
  _clrbit_,CLRBIT,
  _tstbit_,TSTBIT,
  _delay_,DELAY,
  _end_, END,
  _norm_,NORM,
  _rem_,REM,
  _empty_, END  /* mark end of table */
};

void cr_lf(void)
{
  putchar(0x0D);
  putchar(0x0A);
}

void stop_executive(void)
{
  while(1);
}

char get_var(char opt)
{
char var;
  get_token();
  var = toupper(*bas_token)-'A'; /* get the input var */
  if(opt)
  {
  get_token();
  if(*bas_token!=',')
                    serror(1);
}
return var;
}

int get_line_number(char __flash  *ptr)
{
int cntr = 0;
char __flash *wp;
  wp = (__flash char*)( START_BASIC_PROGRAM);
  for(;;)
  {
   if(wp == ptr)
    break;
      if(*wp == 0x0D)
                cntr++;
  wp++;
  }

return cntr;
}


void norm(void)
{
char var;
long  long_param1;
short short_param;

 var = get_var(1);
 get_exp(&short_param);
 long_param1 = bas_variables[var];
 long_param1 = long_param1 * short_param;
   get_token();
  if(*bas_token!=',')
                    serror(1);
 get_exp(&short_param);
 bas_variables[var] = long_param1 / short_param;
}


void interpreter_main(void)
{

//  int_wrk = &interpreter_work_area;
  bas_prog = (__flash char*)( START_BASIC_PROGRAM);
  scan_labels(); /* find the labels in the program */
  bas_ftos = 0; /* initialize the FOR stack index */
  bas_gtos = 0; /* initialize the GOSUB stack index */
  do {
    keep_alive_timer = 250;
    LED5_ON();
    bas_token_type = get_token();
    /* check for assignment statement */

    if(bas_token_type==VARIABLE) {
      putback(); /* return the var to the input stream */
      assignment(); /* must be assignment statement */
    }
    else /* is command */
      {
      switch(bas_tok) {
        case REM:
           find_eol();
	  break;
        case NORM:
           norm();
	  break;
        case SETBIT:
           set_bit();
	  break;
        case CLRBIT:
           clr_bit();
	  break;
        case DELAY:
           delay_();
	  break;
        case TSTBIT:
           tst_bit();
	  break;
        case ADC:
           adc();
	  break;
        case DAC:
           dac();
	  break;
        case PRINT:
	  print();
  	  break;
        case GOTO:
	  exec_goto();
	  break;
	case IF:
	  exec_if();
	  break;
	case FOR:
	  exec_for();
	  break;
	case NEXT:
	  next();
	  break;
  	case INPUT:
	  input_();
	  break;
        case GOSUB:
	  gosub();
	  break;
	case RETURN:
	  greturn();
	  break;
        case END:
    stop_executive();
      }
     }
  } while (bas_tok != FINISHED);
stop_executive();
}


/* Assign a variable a value. */
void assignment(void)
{
  ARITH_BASE var, value;

  /* get the variable name */
  get_token();
  if(!isalpha(*bas_token)) {
    serror(4);
    return;
  }

  var = toupper(*bas_token)-'A';

  /* get the equals sign */
  get_token();
  if(*bas_token!='=') {
    serror(3);
    return;
  }

  /* get the value to assign to var */
  get_exp(&value);

  /* assign the value */
  bas_variables[var] = value;
}

/* Execute a simple version of the BASIC PRINT statement */
void print()
{
  ARITH_BASE answer;
  int len=0, spaces;
  char last_delim;
  char first_char;

  do {
    get_token(); /* get next list item */
    if(bas_tok==EOL || bas_tok==FINISHED)
       {
        bas_tok = bas_tok;
        break;
       }
    if(bas_token_type==QUOTE) { /* is string */
      first_char = *bas_token;
      if(first_char != '$')
      {
      putstr(bas_token);
      len += strlen(bas_token);
      }
      get_token();

    }
    else { /* is expression */
      putback();
      get_exp(&answer);
      get_token();
      if(first_char != '$')
      len += putstr(int_to_str(answer));
      else
      {
       len++;
       first_char = 0x00;
         putchar(answer);
      }
    }
    last_delim = *bas_token;

    if(*bas_token==',') {
      /* compute number of spaces to move to next tab */
      spaces = 8 - (len % 8);
      len += spaces; /* add in the tabbing position */
      while(spaces) {

  putchar(' ');



        spaces--;
      }
    }
    else if(*bas_token==';') /* do nothing */;
    else if(bas_tok!=EOL && bas_tok!=FINISHED) serror(0);
  } while (*bas_token==';' || *bas_token==',');

  if(bas_tok==EOL || bas_tok==FINISHED) {
    if(last_delim != ';' && last_delim!=',')
                                         cr_lf();
  }
  else serror(0); /* error is not , or ; */

}

/* Find all labels. */
void scan_labels()
{
  int addr;
 char __flash *temp;

  label_init();  /* zero all labels */
  temp = bas_prog;   /* save pointer to top of program */

//* if the first token in the file is a label */
//  get_token();
//  if(bas_token_type==NUMBER) {
//    strcpy(bas_label_table[0].name,bas_token);
//    bas_label_table[0].p= bas_prog;
//  }
//  if(bas_token_type!=DELIMITER)
//     find_eol();
  do {
    get_token();
    if(bas_token_type==NUMBER) {
      addr = get_next_label(bas_token);
      if(addr==-1 || addr==-2) {
          (addr==-1) ?serror(5):serror(6);
      }
      strcpy(bas_label_table[addr].name,bas_token);
      bas_label_table[addr].p = bas_prog;  /* current point in program */
    }
    /* if not on a blank line, find next line */
    if(bas_tok!=EOL) find_eol();
  } while(bas_tok!=FINISHED);
  bas_prog = temp;  /* restore to original */
}
//
/* Find the start of the next line. */
void find_eol()
{
  while(*bas_prog!='\n'  && *bas_prog!='\0') ++bas_prog;
  if(*bas_prog) bas_prog++;
}
 //
/* Return index of next free position in label array.
   A -1 is returned if the array is full.
   A -2 is returned when duplicate label is found.
*/
int get_next_label(char *s)
{
  register int t;

  for(t=0;t<NUM_LAB;++t) {
    if(bas_label_table[t].name[0]==0) return t;
    if(!strcmp(bas_label_table[t].name,s)) return -2; /* dup */
  }

  return -1;
}

/* Find location of given label.  A null is returned if
   label is not found; otherwise a pointer to the position
   of the label is returned.
*/
char __flash *find_label(char *s)
{
  register int t;

  for(t=0; t<NUM_LAB; ++t)
    if(!strcmp(bas_label_table[t].name,s))
      return bas_label_table[t].p;
  return '\0'; /* error condition */
}

/* Execute a GOTO statement. */
void exec_goto()
{

  char __flash *loc;

  get_token(); /* get label to go to */
  /* find the location of the label */
  loc = find_label(bas_token);
  if(loc=='\0')
    serror(7); /* label not defined */

  else bas_prog=loc;  /* start program running at that loc */
}

/* Initialize the array that holds the labels.
   By convention, a null label name indicates that
   array position is unused.
*/
void label_init()
{
  register int t;

  for(t=0; t<NUM_LAB; ++t) bas_label_table[t].name[0]='\0';
}

/* Execute an IF statement. */
void exec_if()
{
  ARITH_BASE x , y, cond;
  char op;

  get_exp(&x); /* get left expression */

  get_token(); /* get the operator */
  if(!strchr("=<>", *bas_token)) {
    serror(0); /* not a legal operator */
    return;
  }
  op=*bas_token;

  get_exp(&y); /* get right expression */

  /* determine the outcome */
  cond = 0;
  switch(op) {
    case '<':
      if(x<y) cond=1;
      break;
    case '>':
      if(x>y) cond=1;
      break;
    case '=':
      if(x==y) cond=1;
      break;
  }
  if(cond) { /* is true so process target of IF */
    get_token();
    if(bas_tok!=THEN) {
      serror(8);
      return;
    }/* else program execution starts on next line */
  }
  else find_eol(); /* find start of next line */
}

/* Execute a FOR loop. */
void exec_for()
{
  FOR_STACK i;
  ARITH_BASE value;

  get_token(); /* read the control variable */
  if(!isalpha(*bas_token)) {
    serror(4);
    return;
  }

  i.var=toupper(*bas_token)-'A'; /* save its index */

  get_token(); /* read the equals sign */
  if(*bas_token!='=') {
    serror(3);
    return;
  }

  get_exp(&value); /* get initial value */

  bas_variables[i.var]=value;

  get_token();
  if(bas_tok!=TO) serror(9); /* read and discard the TO */

  get_exp(&i.target); /* get target value */

  /* if loop can execute at least once, push info on stack */
  if(value>=bas_variables[i.var]) {
    i.loc = bas_prog;
    fpush(i);
  }
  else  /* otherwise, skip loop code altogether */
    while(bas_tok!=NEXT) get_token();
}

/* Execute a NEXT statement. */
void next()
{
  FOR_STACK i;

  i = fpop(); /* read the loop info */

  bas_variables[i.var]++; /* increment control variable */
  if(bas_variables[i.var]>i.target) return;  /* all done */
  fpush(i);  /* otherwise, restore the info */
  bas_prog = i.loc;  /* loop */
}

/* Push function for the FOR stack. */
void fpush(FOR_STACK i)

{
   if(bas_ftos>FOR_NEST)
    serror(10);

  bas_fstack[bas_ftos]=i;
  bas_ftos++;
}

FOR_STACK fpop(void)
{
  bas_ftos--;
  if(bas_ftos<0) serror(11);
  return(bas_fstack[bas_ftos]);
}

/* Execute a simple form of the BASIC INPUT command */
void input_()
{
  char var;
  int i;
  char *wp, sym;
  get_token(); /* see if prompt string is present */
  if(bas_token_type==QUOTE) {
    sym = *bas_token;
    if(sym != '$')
    putstr(bas_token); /* if so, print it and check for comma */
    get_token();
    if(*bas_token!=',') serror(1);
    get_token();
  }
    else
        putstr("? "); /* otherwise, prompt with / */

    var = toupper(*bas_token)-'A'; /* get the input var */

  if(sym=='$')
   {
       i = getchar();
   }
  else
   {
    wp = input_line();
    i = atoi(wp);
       cr_lf();
   }
  bas_variables[var] = i; /* store it */

}

/* Execute a GOSUB command. */
void gosub(void)
{
  char __flash *loc;

  get_token();
  /* find the label to call */
  loc = find_label(bas_token);
  if(loc=='\0')
    serror(7); /* label not defined */
  else {
    gpush(bas_prog); /* save place to return to */
    bas_prog = loc;  /* start program running at that loc */
  }
}

/* Return from GOSUB. */
void greturn()
{
   bas_prog = gpop();
}

/* GOSUB stack push function. */
void gpush(char __flash *s)
{
  bas_gtos++;

  if(bas_gtos==SUB_NEST) {
    serror(12);
    return;
  }

  bas_gstack[bas_gtos]=s;

}

/* GOSUB stack pop function. */
char __flash *gpop(void)
{
  if(bas_gtos==0) {
    serror(13);
    return 0;
  }

  return(bas_gstack[bas_gtos--]);
}

/* Entry point into parser. */
void get_exp(ARITH_BASE *result)

{
  get_token();
  if(!*bas_token) {
    serror(2);
    return;
  }
  level2(result);
  putback(); /* return last token read to input stream */
}
__flash char _fmt_s_n_[] = "\r\nERROR:";

void serror(int error)
{
  putstr_P(_fmt_s_n_);
  putstr(int_to_str(error));

  putchar(',');

  putstr(int_to_str(get_line_number(bas_prog)+1));

  cr_lf();

   stop_executive();

}

/* Get a token. */
int get_token(void)
{

  char *temp;

  bas_token_type=0; bas_tok=0;
  temp=bas_token;
  if(*bas_prog=='\0') { /* end of file */
    *bas_token=0;
    bas_tok = FINISHED;
    return(bas_token_type=DELIMITER);
  }

  while(iswhite(*bas_prog)) ++bas_prog;  /* skip over white space */

  if(*bas_prog=='\r') { /* crlf */
    ++bas_prog; ++bas_prog;
    bas_tok = EOL;
    *bas_token='\r';
    bas_token[1]='\n'; bas_token[2]=0;

    return (bas_token_type = DELIMITER);
  }

  if(strchr("&|+-*/%=;(),><", *bas_prog)){ /* delimiter */
    *temp=*bas_prog;
    bas_prog++; /* advance to next position */
    temp++;
    *temp=0;
    return (bas_token_type=DELIMITER);
  }

  if(*bas_prog=='"') { /* quoted string */
    bas_prog++;
    while(*bas_prog!='"'&& *bas_prog!='\r') *temp++=*bas_prog++;
    if(*bas_prog=='\r') serror(1);
    bas_prog++;*temp=0;
    return(bas_token_type=QUOTE);
  }

  if(isdigit(*bas_prog)) { /* number */
    while(!isdelim(*bas_prog)) *temp++=*bas_prog++;
    *temp = '\0';
    return(bas_token_type = NUMBER);
  }

  if(isalpha(*bas_prog)) { /* var or command */
    while(!isdelim(*bas_prog)) *temp++=*bas_prog++;
    bas_token_type=STRING;
  }

  *temp = '\0';

  /* see if a string is a command or a variable */
  if(bas_token_type==STRING) {
    bas_tok=look_up(bas_token); /* convert to internal rep */
    if(!bas_tok) bas_token_type = VARIABLE;
    else bas_token_type = COMMAND; /* is a command */
  }
  return bas_token_type;
}



/* Return a token to input stream. */
void putback()
{

  char *t;

  t = bas_token;
  for(; *t; t++) bas_prog--;
}

/* Look up a a token's internal representation in the
   token table.
*/

char cmp_P(char __flash *s,char __flash *p)
{
return 1;
}

char look_up(char *s)
{
  register int i;
  char *p;

  /* convert to lowercase */
  p = s;
  while(*p){ *p = tolower(*p); p++; }

  /* see if token is in table */
  for(i=0; *table[i].command; i++)
      if(!strcmp_P(s, table[i].command))
                     return table[i].tok;
  return 0; /* unknown command */
}

/* Return true if c is a delimiter. */
int isdelim(char c)
{
  if(strchr(" ;,&|+-<>/*%=()", c) || c==9 || c=='\r' || c==0)
    return 1;
  return 0;
}

/* Return 1 if c is space or tab. */
int iswhite(char c)
{
  if(c==' ' || c=='\t') return 1;
  else return 0;
}



/*  Add or subtract two terms. */
void level2(ARITH_BASE *result)

{
  register char  op;
  ARITH_BASE hold;

  level3(result);
  while((op = *bas_token) == '+' || op == '-') {
    get_token();
    level3(&hold);
    arith(op, result, &hold);
  }
}

/* Multiply or divide two factors. */
void level3(ARITH_BASE *result)
{
  register char  op;
  ARITH_BASE hold;

  level4(result);
  while((op = *bas_token) == '*' || op == '/' || op == '%' || op == '&' || op == '|') {
    get_token();
    level4(&hold);
    arith(op, result, &hold);
  }
}

/* Process integer exponent. */
void level4(ARITH_BASE *result)
{
//  ARITH_BASE hold;

  level5(result);
//  if(*bas_token== '^') {
//    get_token();
//    level4(&hold);
//    arith('^', result, &hold);
//  }
}

/* Is a unary + or -. */
void level5(ARITH_BASE *result)
{
  register char  op;

  op = 0;
  if((bas_token_type==DELIMITER) && *bas_token=='+' || *bas_token=='-') {
    op = *bas_token;
    get_token();
  }
  level6(result);
  if(op)
    unary(op, result);
}

/* Process parenthesized expression. */
void level6(ARITH_BASE *result)
{
  if((*bas_token == '(') && (bas_token_type == DELIMITER)) {
    get_token();
    level2(result);
    if(*bas_token != ')')
      serror(1);
    get_token();
  }
  else
    primitive(result);
}

/* Find value of number or variable. */
void primitive(ARITH_BASE *result)
{

  switch(bas_token_type) {
  case VARIABLE:
    *result = find_var(bas_token);
    get_token();
    return;
  case NUMBER:
    *result = atoi(bas_token);
    get_token();
    return;
  default:
    serror(0);
  }
}

/* Perform the specified arithmetic. */
void arith(char o, ARITH_BASE *r, ARITH_BASE *h)
{
  register int t;//, ex;

  switch(o) {
    case '&':
      *r = *r & *h;
      break;
    case '|':
      *r = *r | *h;
      break;
    case '-':
      *r = *r-*h;
      break;
    case '+':
      *r = *r+*h;
      break;
    case '*':
      *r = *r * *h;
      break;
    case '/':
      *r = (*r)/(*h);
      break;
    case '%':
      t = (*r)/(*h);
      *r = *r-(t*(*h));
      break;
//    case '^':
//      ex = *r;
//      if(*h==0) {
//        *r = 1;
//        break;
//      }
//      for(t=*h-1; t>0; --t) *r = (*r) * ex;
//      break;
  }
}

/* Reverse the sign. */
void unary(char o, ARITH_BASE *r)

{
  if(o=='-') *r = -(*r);
}

/* Find the value of a variable. */
int find_var(char *s)
{
  if(!isalpha(*s)){
    serror(4); /* not a variable */
    return 0;
  }
  return bas_variables[toupper(*bas_token)-'A'];
}


